"""Provides access to pixel data."""

import io
from base64 import b64encode
from contextlib import nullcontext
from pathlib import Path
from typing import ContextManager, Iterable, Iterator, Literal, Tuple

from PIL import Image as PILImage

from textual_kitty.utils import grouped


def ensure_image(image: str | Path | PILImage.Image) -> ContextManager[PILImage.Image]:
    """Ensures value to be an `PIL.Image.Image`.

    This function accepts either a str or `pathlib.Path` of a path to an image file or a `PIL.Image.Image` instance.
    It returns a context manager that either just provides the passed `PIL.Image.Image` instance or opens the path
    with `PIL` and provides the result.

    Args:
        image: Path to an image file or `PIL.Image.Image` instance.

    Returns:
        A context manager providing a `PIL.Image.Image` instance.
    """
    if isinstance(image, PILImage.Image):
        return nullcontext(image)
    return PILImage.open(image)


class PixelMeta:
    """Provides access to meta information of an image from a path or `PIL.Image.Image` instance."""

    def __init__(self, image: str | Path | PILImage.Image) -> None:
        """Initializes a PixelMeta.

        Args:
            image: Path to an image file or `PIL.Image.Image` instance with the image data to render.
        """
        with ensure_image(image) as opened_image:
            self.width = opened_image.width
            self.height = opened_image.height


class PixelData:
    """Provides access to pixel data from a path or `PIL.Image.Image` instance."""

    def __init__(self, image: str | Path | PILImage.Image, mode: Literal["grayscale", "rgb"] | None = None) -> None:
        """Initializes a PixelData.

        Args:
            image: Path to an image file or `PIL.Image.Image` instance with the image data to render.
            mode: Mode to convert the image into or `None` to keep the original mode.
        """
        with ensure_image(image) as opened_image:
            self._image = opened_image.copy()

        if mode == "grayscale":
            self._image = self._image.convert("L")
        elif mode == "rgb":
            self._image = self._image.convert("RGB")

    @property
    def width(self) -> int:
        """The image's width."""
        return self._image.width

    @property
    def height(self) -> int:
        """The image's height."""
        return self._image.height

    def scaled(self, width: int, height: int) -> "PixelData":
        """Returns a scaled copy of this PixelData.

        Args:
            width: Width to scale to.
            height: Height to scale to.

        Returns:
            A new `PixelData` instance of the same image, scaled to (width, height).
        """
        scaled_image = self._image.resize((width, height))
        return PixelData(scaled_image)

    def to_base64(self) -> str:
        """Return the pixel data as base64 encoded PNG.

        Returns:
        The image data, as base64 encoded PNG.
        """
        image_buffer = io.BytesIO()
        self._image.save(image_buffer, format="png", compress_level=2)
        return b64encode(image_buffer.getvalue()).decode("ascii")

    def __iter__(self) -> Iterator[Iterable[int | Tuple[int, ...]]]:
        """Iterate the pixel data.

        Returns an `Iterator` over rows of the image.
        The rows are `Iterable`s of pixel values.

        Example:
            Iterate the image data::

            for row in PixelData("image.png"):
                for pixel in row:
                    ...  # Do something with the pixel here

        Returns:
            An `Iterator` over `Iterable`s of pixel values.
        """
        data = self._image.getdata()
        yield from grouped(data, self._image.width)



#!/usr/bin/env python

"""Showcase textual-kitty Renderables/Widgets."""

from argparse import ArgumentParser
from pathlib import Path
from typing import cast

from PIL import Image as PILImage

from textual_kitty.renderable import (
    HalfcellImage as HalfcellRenderable,
)
from textual_kitty.renderable import (
    Image as AutoRenderable,
)
from textual_kitty.renderable import (
    TGPImage as TGPRenderable,
)
from textual_kitty.renderable import (
    UnicodeImage as UnicodeRenderable,
)
from textual_kitty.widget import HalfcellImage as HalfcellWidget
from textual_kitty.widget import Image as AutoWidget
from textual_kitty.widget import TGPImage as TGPWidget
from textual_kitty.widget import UnicodeImage as UnicodeWidget

TEST_IMAGE = Path(__file__).parent / "gracehopper.jpg"

RENDERING_PROTOCOLS = {
    "auto": (AutoRenderable, AutoWidget),
    "tgp": (TGPRenderable, TGPWidget),
    "halfcell": (HalfcellRenderable, HalfcellWidget),
    "unicode": (UnicodeRenderable, UnicodeWidget),
}


def demo_rich(protocol: str) -> None:
    """Showcase textual-kitty's Rich renderables for images."""
    from rich.console import Console
    from rich.style import Style
    from rich.table import Table

    Image = RENDERING_PROTOCOLS[protocol][0]

    console = Console()

    console.print("Image, no size given:", style=Style(bold=True))
    console.print(Image(TEST_IMAGE))

    console.print("Image, 100% width:", style=Style(bold=True))
    console.print(Image(TEST_IMAGE, width="100%"))

    console.print("Image, 40 cells width, height auto:", style=Style(bold=True))
    console.print(Image(TEST_IMAGE, width=40, height="auto"))

    console.print("Image inside Table:", style=Style(bold=True))
    table = Table("Name", "Birthday", "Picture")
    table.add_row("Grace Hopper", "December 9, 1906", Image(TEST_IMAGE))
    console.print(table)


def demo_textual(protocol: str) -> None:
    """Showcase textual-kitty's Textual Widget for images."""
    from textual import on
    from textual.app import App, ComposeResult
    from textual.containers import Container, Horizontal, HorizontalScroll, ScrollableContainer
    from textual.css.scalar import Scalar
    from textual.widgets import Button, Footer, Header, TabbedContent, TabPane

    Image = RENDERING_PROTOCOLS[protocol][1]

    class DemoApp(App[None]):
        CSS = """
        DemoApp {
            #sizing {
                layout: grid;
                grid-size: 3 2;
            }

            #sizing > * {
                border: round gray;
                align: center middle;
            }

            #zoom {
                border: round gray;
            }

            #zoom > * {
                align: center middle;
            }

            #zoom-in {
                margin-right: 1;
                width: 1fr;
            }

            #zoom-out {
                margin-left: 1;
                width: 1fr;
            }

            #many-container {
                align: center middle;
            }

            #many-container > Image {
                border: round gray;
            }

            .width-auto {
                width: auto;
            }

            .height-auto {
                height: auto;
            }

            .width-15 {
                width: 15;
            }

            .height-50pct {
                height: 50%;
            }

            .width-100pct {
                width: 100%;
            }
        }
        """

        def compose(self) -> ComposeResult:
            yield Header()
            with TabbedContent():
                with TabPane("Sizing", id="sizing"):
                    with Container() as c:
                        c.border_title = "width: none; height: none;"
                        yield Image(TEST_IMAGE, classes="width-50-pc")
                    with Container() as c:
                        c.border_title = "width: auto; height: none;"
                        yield Image(TEST_IMAGE, classes="width-auto")
                    with Container() as c:
                        c.border_title = "width: none; height: auto;"
                        yield Image(TEST_IMAGE, classes="height-auto")
                    with Container() as c:
                        c.border_title = "width: auto; height: auto;"
                        yield Image(TEST_IMAGE, classes="width-auto height-auto")
                    with Container() as c:
                        c.border_title = "width: 15; height: auto;"
                        yield Image(TEST_IMAGE, classes="width-15 height-auto")
                    with Container() as c:
                        c.border_title = "width: auto; height: 50%;"
                        yield Image(TEST_IMAGE, classes="width-auto height-50pct")
                with TabPane("Zoom and Scroll", id="zoom"):
                    with ScrollableContainer():
                        yield Image(TEST_IMAGE, id="zoomable", classes="width-100pct height-auto")
                    with Horizontal(classes="height-auto"):
                        yield Button("+", id="zoom-in")
                        yield Button("-", id="zoom-out")
                with TabPane("Many Large Images", id="many"):
                    with PILImage.open(TEST_IMAGE) as large_image:
                        large_image = large_image.resize((large_image.width * 4, large_image.height * 4))
                    with HorizontalScroll(id="many-container"):
                        for _ in range(100):
                            yield Image(large_image, classes="width-auto height-50pct")
            yield Footer()

        @on(Button.Pressed, "#zoom-in")
        def zoom_in(self, event: Button.Pressed) -> None:
            widget = self.query_one("#zoomable", Image)
            widget.styles.width = int(cast(Scalar, widget.styles.width).value * 1.1)

        @on(Button.Pressed, "#zoom-out")
        def zoom_out(self, event: Button.Pressed) -> None:
            widget = self.query_one("#zoomable", Image)
            widget.styles.width = int(cast(Scalar, widget.styles.width).value / 1.1)

    DemoApp().run()


def run() -> None:
    """Run the textual_kitty demo."""
    parser = ArgumentParser(description="Demo the capabilities of textual-kitty")
    parser.add_argument("mode", choices=["rich", "textual"], nargs="?", default="rich")
    parser.add_argument("-p", "--protocol", choices=RENDERING_PROTOCOLS.keys(), default="auto")
    arguments = parser.parse_args()

    if arguments.mode == "rich":
        demo_rich(arguments.protocol)
    elif arguments.mode == "textual":
        demo_textual(arguments.protocol)


if __name__ == "__main__":
    run()



from pytest import raises

from tests.data import TERMINAL_SIZES
from textual_kitty.geometry import ImageSize


def test_image_size_validation() -> None:
    ImageSize(256, 256, None, None).validate()
    ImageSize(256, 256, 128, 128).validate()
    ImageSize(256, 256, "10%", "10%").validate()
    ImageSize(256, 256, "auto", "auto").validate()

    with raises(ValueError):
        ImageSize(256, 256, "-10%", "-10%").validate()

    with raises(ValueError):
        ImageSize(256, 256, "10", "10").validate()

    with raises(ValueError):
        ImageSize(256, 256, "xx%", "xx%").validate()


def test_image_size_cell_size_calculation() -> None:
    assert ImageSize(0, 0, None, None).get_cell_size(100, 100, TERMINAL_SIZES) == (0, 0)
    assert ImageSize(256, 256, None, None).get_cell_size(100, 100, TERMINAL_SIZES) == (32, 16)
    assert ImageSize(256, 256, "50%", "50%").get_cell_size(100, 100, TERMINAL_SIZES) == (50, 50)
    assert ImageSize(256, 256, "auto", "auto").get_cell_size(100, 100, TERMINAL_SIZES) == (100, 50)
    assert ImageSize(256, 256, "50%", "auto").get_cell_size(100, 100, TERMINAL_SIZES) == (50, 25)
    assert ImageSize(256, 256, 50, "auto").get_cell_size(100, 100, TERMINAL_SIZES) == (50, 25)
    assert ImageSize(256, 256, "auto", "50%").get_cell_size(100, 100, TERMINAL_SIZES) == (100, 50)
    assert ImageSize(256, 256, "auto", 50).get_cell_size(100, 100, TERMINAL_SIZES) == (100, 50)
    assert ImageSize(256, 256, "auto", None).get_cell_size(32, 32, TERMINAL_SIZES) == (32, 16)
    assert ImageSize(256, 256, None, "auto").get_cell_size(32, 32, TERMINAL_SIZES) == (32, 32)
    assert ImageSize(12, 512, "auto", "auto").get_cell_size(32, 32, TERMINAL_SIZES) == (2, 32)


def test_image_size_pixel_size_calculation() -> None:
    assert ImageSize(256, 256, None, None).get_pixel_size(100, 100, TERMINAL_SIZES) == (256, 256)



import io
from typing import Any, AsyncContextManager, cast
from unittest.mock import patch

from rich.console import Console
from textual.app import App
from textual.pilot import Pilot
from textual.widgets import Button

from textual_kitty import demo


def test_arg_parsing() -> None:
    with patch("sys.argv", ["demo", "rich"]):
        with patch("textual_kitty.demo.demo_rich") as demo_rich:
            demo.run()
    assert demo_rich.called

    with patch("sys.argv", ["demo", "textual"]):
        with patch("textual_kitty.demo.demo_textual") as demo_textual:
            demo.run()
    assert demo_textual.called


def test_rich() -> None:
    stdout = io.StringIO()
    with patch.object(Console, "file", stdout):
        demo.demo_rich("auto")
    assert "█▒▒░░░░░░░░░░" in stdout.getvalue()


async def test_textual() -> None:
    # This is incredibly hacky. But is seems to work.
    awaitable = None

    def run_wrapper(app: App[Any]) -> None:
        nonlocal awaitable
        awaitable = app.run_test()

    # This doesn't actually test much, but at least we run the code.
    with patch.object(App, "run", run_wrapper):
        demo.demo_textual("auto")
        async with cast(AsyncContextManager[Pilot[Any]], awaitable) as pilot:
            pilot.app.query_one("#zoom-in", Button).press()
            pilot.app.query_one("#zoom-out", Button).press()
            await pilot.pause()



from textual_kitty.utils import clamp, grouped


def test_grouped() -> None:
    assert list(grouped([1, 2, 3, 4, 5, 6], 2)) == [(1, 2), (3, 4), (5, 6)]


def test_clamp() -> None:
    assert clamp(1, 50, 100) == 50
    assert clamp(150, 50, 100) == 100
    assert clamp(75, 50, 100) == 75
    assert clamp(50, 50, 100) == 50
    assert clamp(100, 50, 100) == 100



from array import array
from typing import IO
from unittest.mock import patch

from pytest import raises

from tests.data import TERMINAL_SIZES
from textual_kitty.terminal import TerminalError, TerminalSizes, capture_terminal_response, get_terminal_sizes


def test_get_terminal_sizes_on_tty_success() -> None:
    def ioctl(fd: IO[bytes], request: int, buf: array[int]) -> None:
        buf[0], buf[1], buf[2], buf[3], _, _ = TERMINAL_SIZES

    with patch("sys.__stdout__.isatty", return_value=True):
        with patch("textual_kitty.terminal.ioctl", side_effect=ioctl):
            assert get_terminal_sizes() == TERMINAL_SIZES


def test_get_terminal_sizes_on_tty_no_screen_size() -> None:
    terminal_sizes = TerminalSizes(
        rows=TERMINAL_SIZES.rows,
        columns=TERMINAL_SIZES.columns,
        screen_width=0,
        screen_height=0,
        cell_width=TERMINAL_SIZES.cell_width,
        cell_height=TERMINAL_SIZES.cell_height,
    )

    def ioctl(fd: IO[bytes], request: int, buf: array[int]) -> None:
        buf[0], buf[1], buf[2], buf[3], _, _ = terminal_sizes

    with patch("sys.__stdout__.isatty", return_value=True):
        with patch("textual_kitty.terminal.ioctl", side_effect=ioctl):
            assert get_terminal_sizes() == terminal_sizes


def test_get_terminal_sizes_on_tty_failure() -> None:
    with patch("sys.__stdout__.isatty", return_value=True):
        with patch("textual_kitty.terminal.ioctl", side_effect=OSError()):
            with raises(TerminalError):
                get_terminal_sizes()


def test_get_terminal_sizes_stdout_closed() -> None:
    with patch("sys.__stdout__", None):
        with raises(TerminalError):
            get_terminal_sizes()


def test_get_terminal_sizes_stdout_not_a_tty() -> None:
    with patch("sys.__stdout__.isatty", return_value=False):
        term_size = get_terminal_sizes()

    assert term_size.screen_width == 0
    assert term_size.screen_height == 0


def test_capture_terminal_response() -> None:
    with patch("sys.__stdin__", None):
        with raises(TerminalError):
            with capture_terminal_response("[S]", "[E]") as response:
                pass

    with patch("sys.__stdin__") as stdin, patch("termios.tcgetattr"), patch("termios.tcsetattr"), patch(
        "tty.setcbreak"
    ), patch("textual_kitty.terminal.select") as select, patch("os.read") as read:
        stdin.buffer.fileno.return_value = 42
        select.return_value = [[42], [], []]
        read.side_effect = [c.to_bytes() for c in b"[S]message[E]"]

        with capture_terminal_response("[S]", "[E]") as response:
            pass

        assert response.sequence == "[S]message[E]"

    with patch("sys.__stdin__") as stdin, patch("termios.tcgetattr"), patch("termios.tcsetattr"), patch(
        "tty.setcbreak"
    ), patch("textual_kitty.terminal.select") as select:
        stdin.buffer.fileno.return_value = 42
        select.return_value = [[], [], []]

        with raises(TimeoutError):
            with capture_terminal_response("[S]", "[E]") as response:
                pass

    with patch("sys.__stdin__") as stdin, patch("termios.tcgetattr"), patch("termios.tcsetattr"), patch(
        "tty.setcbreak"
    ), patch("textual_kitty.terminal.select") as select, patch("os.read") as read:
        stdin.buffer.fileno.return_value = 42
        select.return_value = [[42], [], []]
        read.side_effect = [c.to_bytes() for c in b"not the expected message"]

        with raises(TerminalError):
            with capture_terminal_response("[S]", "[E]") as response:
                pass



from base64 import b64decode

from PIL import Image as PILImage

from tests.data import TEST_IMAGE
from textual_kitty.pixeldata import PixelData, PixelMeta, ensure_image


def test_ensure_image() -> None:
    with ensure_image(TEST_IMAGE) as image:
        assert isinstance(image, PILImage.Image)

    with PILImage.open(TEST_IMAGE) as opened_image:
        with ensure_image(opened_image) as image:
            assert isinstance(image, PILImage.Image)
            assert image is opened_image


def test_pixel_meta() -> None:
    with PILImage.open(TEST_IMAGE) as image:
        meta = PixelMeta(image)
        assert meta.width == image.width
        assert meta.height == image.height


def test_pixel_data_mode() -> None:
    with PILImage.open(TEST_IMAGE) as image:
        assert PixelData(image, mode="grayscale")._image.mode == "L"
        assert PixelData(image, mode="rgb")._image.mode == "RGB"


def test_pixel_data_size() -> None:
    with PILImage.open(TEST_IMAGE) as image:
        data = PixelData(image)
        assert data.width == image.width
        assert data.height == image.height


def test_pixel_data_scale() -> None:
    data = PixelData(TEST_IMAGE).scaled(32, 32)
    assert data._image.width == 32
    assert data._image.height == 32


def test_pixel_data_to_base64() -> None:
    data = PixelData(TEST_IMAGE).to_base64()
    b64decode(data)  # must not raise


def test_pixel_data_iteration() -> None:
    data = PixelData(TEST_IMAGE).scaled(8, 8)
    assert len(list(data)) == 8
    assert all(len(list(row)) == 8 for row in data)



"""Functionality to interact with the terminal."""

import logging
import os
import sys
import termios
import tty
from array import array
from contextlib import contextmanager
from fcntl import ioctl
from select import select
from types import SimpleNamespace
from typing import Iterator, NamedTuple

logger = logging.getLogger(__name__)


class TerminalError(Exception):
    """Error thrown on failing terminal operations."""

    pass


class TerminalSizes(NamedTuple):
    """Size of several terminal features."""

    rows: int
    """Width of the terminal in cells."""
    columns: int
    """Height of the terminal in cells."""
    screen_width: int
    """Width of the terminal in pixels."""
    screen_height: int
    """Height of the terminal in pixels."""
    cell_width: int
    """Width of a terminal cell in pixels."""
    cell_height: int
    """Height of a terminal cell in pixels."""


def get_terminal_sizes() -> TerminalSizes:
    """Get size information from the terminal.

    Returns:
        The size information

    """
    if not sys.__stdout__:
        raise TerminalError("stdout is closed")

    if sys.__stdout__.isatty():
        try:
            buf = array("H", [0, 0, 0, 0])
            ioctl(sys.__stdout__, termios.TIOCGWINSZ, buf)

            rows, columns, screen_width, screen_height = buf

            # If we can't get the cell size (not all terminals support this) we're just assuming some hopefully reasonable numbers.
            # Does this suck? Yes. Is there something we can do about it? Not really.
            if not screen_width or not screen_height:
                logger.warning("Terminal doesn't support getting pixel size; falling back to assumptions")

        except OSError as e:
            raise TerminalError("Unsupported terminal") from e

    else:
        logger.debug("Not connected to a terminal, assuming reasonable size information")
        rows = 24
        columns = 80
        screen_width = 0
        screen_height = 0

    cell_width = int(screen_width / columns) or 8
    cell_height = int(screen_height / rows) or 16

    return TerminalSizes(
        rows=rows,
        columns=columns,
        screen_width=screen_width,
        screen_height=screen_height,
        cell_width=cell_width,
        cell_height=cell_height,
    )


@contextmanager
def capture_terminal_response(
    start_marker: str, end_marker: str, timeout: float | None = None
) -> Iterator[SimpleNamespace]:
    """Captures a terminal response.

    Captures the terminal's response to an escape sequence.
    This is a bit flaky -- keystrokes during reading the response can lead to false answers.
    Additionally, when the terminal does *not* doesn't send an answer, the first character
    of stdin may get lost as this function reads it to determine if it is the response.
    Anyway, as this is improbable to happen, it should be fine.

    Please not this function will not work anymore once Textual is started. Textual runs a threads to read stdin
    and will grab the response.

    Args:
        start_marker: The start sequence of the expected response
        end_marker: The end sequence of the expected response
        timeout: The number of seconds to wait for the response. None to disable timeout.

    Returns:
        The terminal's response
    """
    if not sys.__stdin__:
        raise TerminalError("stdout is closed")

    response = SimpleNamespace(sequence="")

    stdin = sys.__stdin__.buffer.fileno()
    old_term_mode = termios.tcgetattr(stdin)
    tty.setcbreak(stdin, termios.TCSANOW)

    try:
        yield response

        while not response.sequence.endswith(end_marker):
            readable, _, _ = select([stdin], [], [], timeout)
            if not readable:
                raise TimeoutError("Timeout waiting for response")

            response.sequence += os.read(stdin, 1).decode()

            if not response.sequence.startswith(start_marker[: len(response.sequence)]):
                raise TerminalError("Unexpected response from terminal")

    finally:
        termios.tcsetattr(stdin, termios.TCSANOW, old_term_mode)



"""Geometry types and logic."""

from types import NoneType
from typing import NamedTuple, Tuple, cast

from textual_kitty.terminal import TerminalSizes


class ImageSize(NamedTuple):
    """Image size class capable of parsing size specifications and calculating pixel and cell sizes.

    This class expects the size of the source image as `source_pixel_width` and `source_pixel_height`
    and a target width and height as `width` and `height`.
    `width` and `height` can have multiple forms:

    - If set to an integer, they describe the number of cells to be used when rendering the image.
    - If set to a string ending with `%` (e.g. `80%`) they describe the percentage of available render
      space to be used when rending the image.
    - If set to `auto` the image will be scaled to use as much as possible of the available render space
      while keeping the image's aspect ratio.
    - If set to `None` the original images dimension will be used.
    """

    source_pixel_width: int
    source_pixel_height: int
    width: int | str | None
    height: int | str | None

    def validate(self) -> None:
        """Validate the instance's size specifications.

        Raises:
            ValueError: If `width` or `height` are not valid.
        """
        self.validate_value(self.width)
        self.validate_value(self.height)

    @classmethod
    def validate_value(cls, value: int | str | None) -> None:
        """Validate a single size specification.

        Args:
            value: The size specification.

        Raises:
            ValueError: If `width` or `height` are not valid.
        """
        if isinstance(value, (int, NoneType)) or value == "auto":
            return
        if isinstance(value, str) and value.endswith("%") and value[:-1].isdecimal() and int(value[:-1]) >= 0:
            return
        raise ValueError(f"Invalid size: {value}")

    def get_cell_size(self, max_width: int, max_height: int, terminal_sizes: TerminalSizes) -> Tuple[int, int]:
        """Calculates the render size based on original image size, size specification and terminal cell sizes.

        Args:
            max_width: The maximum available render width in cells.
            max_height: The maximum available render height in cells.
            terminal_sizes: Size information of the terminal.

        Returns:
            The render size in cells as tuple of (width, height).
        """
        self.validate()

        if self.source_pixel_height == 0 or self.source_pixel_width == 0:
            return 0, 0

        width = None
        height = None

        # Just use the original image size for dimensions that're `None`
        if self.width is None:
            width = round(self.source_pixel_width / terminal_sizes.cell_width)
        if self.height is None:
            height = round(self.source_pixel_height / terminal_sizes.cell_height)

        # Explicitly given cell sizes
        if isinstance(self.width, int):
            width = self.width
        if isinstance(self.height, int):
            height = self.height

        # Percentage of available space
        if isinstance(self.width, str) and self.width.endswith("%"):
            width = round(max_width / 100 * int(self.width[:-1]))
        if isinstance(self.height, str) and self.height.endswith("%"):
            height = round(max_height / 100 * int(self.height[:-1]))

        # If `width` is `auto` we use the `max_width` if we don't have a height specified,
        # but keep the ratio if `height` was set
        if self.width == "auto" and self.height is not None:
            ratio = self.source_pixel_width / self.source_pixel_height
            scaled_pixel_height: float = (height or max_height) * terminal_sizes.cell_height
            scaled_pixel_width = scaled_pixel_height * ratio
            width = min(round(scaled_pixel_width / terminal_sizes.cell_width), max_width)
        elif self.width == "auto":
            width = max_width

        # Same for height
        if self.height == "auto" and self.width is not None:
            ratio = self.source_pixel_width / self.source_pixel_height
            scaled_pixel_width = cast(int, width) * terminal_sizes.cell_width
            scaled_pixel_height = scaled_pixel_width / ratio
            height = round(scaled_pixel_height / terminal_sizes.cell_height)
        elif self.height == "auto":
            height = max_height

        # By deriving `height` from `width` in the `auto` case, `height` may exceed `max_height`,
        # which shouldn't happen when set to `auto`. Fix this my scaling down both `height` and `width`.`
        if self.width == "auto" and self.height == "auto" and cast(int, height) > max_height:
            height = max_height
            ratio = self.source_pixel_width / self.source_pixel_height
            scaled_pixel_height = height * terminal_sizes.cell_height
            scaled_pixel_width = scaled_pixel_height * ratio
            width = round(scaled_pixel_width / terminal_sizes.cell_width)

        return cast(Tuple[int, int], (width, height))

    def get_pixel_size(self, max_width: int, max_height: int, terminal_sizes: TerminalSizes) -> Tuple[int, int]:
        """Calculates the pixel size to use for rendering based on original image size, size specification and terminal cell sizes.

        Args:
            max_width: The maximum available render width in cells.
            max_height: The maximum available render height in cells.
            terminal_sizes: Size information of the terminal.

        Returns:
            The render size in pixels as tuple of (width, height).
        """
        width, height = self.get_cell_size(max_width, max_height, terminal_sizes)
        return width * terminal_sizes.cell_width, height * terminal_sizes.cell_height



"""Utility functions."""

from typing import Iterable, Iterator, TypeVar

T = TypeVar("T")


def grouped(iterable: Iterable[T], n: int) -> Iterator[Iterable[T]]:
    """Groups values of an iterable into equally sized groups.

    The values of iterable are collected into groups of size `n`.
    An iterator over the resulting groups is returned.

    Example:
        Grouping a list of integers::

            >>> list(grouped([1, 2, 3, 4, 5, 6], 2))
            [(1, 2), (3, 4), (5, 6)]

    Args:
        iterable: Iterable over the values to group
        n: Size of each group

    Returns:
        An iterator over the groups
    """
    return zip(*([iter(iterable)] * n), strict=True)


N = TypeVar("N", bound=int | float)


def clamp(n: N, minimum: N, maximum: N) -> N:
    """Constrains a value to lie between two further values.

    Args:
        n: The value to constrain.
        minimum: The manimum value n may have.
        maximum: The maximum value n may have.

    Returns:
        The constrained value.
    """
    return max(min(n, maximum), minimum)
